/*
COPYRIGHT (C) 1999  Paolo Mantegazza (mantegazza@aero.polimi.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
*/

#define LXRT_MODULE
#include <linux/version.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/wrapper.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_lxrt.h>

#define VERBOSE			// Include printk's
#define ENABLE_SIGNAL		// Enable srq handler to send SIGNALS
#define ENABLE_SIGNAL_TEST	// Enable srq handler test code

#define IS_TASK 0		// Used to identify registered resources
#define IS_SEM  1
#define IS_MBX  2

extern int set_rtai_callback(    void (*fun)(struct task_struct *tsk));
extern void remove_rtai_callback(void (*fun)(struct task_struct *tsk));
extern void rtai_callback( struct task_struct *tsk );
extern int rt_signal_linux_task(RT_TASK *task, int sig);

#define __STR(x) #x
#define STR(x) __STR(x)

/*
 * rtai_lxrt_handler() is the handler for soft interrupt RTAI_LXRT_VECTOR,
 * so the approriate entry in the IDT has been setup to point to it by
 * the call to rt_set_full_intr_vect(...) in init_module.
 *
 * The system call completes by jumping into ret_from_intr() (entry.S).
 * As a result, the Linux scheduler may get called.
 *
*/

static void rtai_lxrt_handler(void)
{
	__asm__ __volatile__ (" \
	pushl %eax;\n\t \
	cld; pushl %es; pushl %ds; pushl %eax; pushl %ebp;\n\t \
	pushl %edi; pushl %esi; pushl %edx; pushl %ecx;\n\t \
	pushl %ebx; pushl %edx; pushl %eax;\n\t \
	movl $" STR(__KERNEL_DS) ",%ebx; movl %bx,%ds; movl %bx,%es");

	__asm__ __volatile__ ("call "SYMBOL_NAME_STR(lxrt_handler));

	__asm__ __volatile__ (" \
	addl $8,%esp; movl %edx,8(%esp); movl %eax,24(%esp);\n\t \
	jmp *"SYMBOL_NAME_STR(rthal + 60));
}

/*     The macro below must be >= the max number of all objects to be 
    registered and shared (objects are: rt_tasks, semaphores, mailboxes).  */
#define MAX_SLOTS  100

static struct {
	unsigned long name;	 // Numerical representation of resource name
	void *adr;		 // Physical rt memory address of resource
	struct task_struct *tsk; // Linux task owner of the resource
	int type;		 // Type of resource
	} lxrt_list[MAX_SLOTS + 1] = {{0, 0, 0, 0}, };
static spinlock_t list_lock = SPIN_LOCK_UNLOCKED;

static inline int registr(unsigned long name, void *adr, int type, struct task_struct *tsk)
{
	unsigned long flags;
	int slot;
/*
 * Register a resource. This allows other programs (RTAI and/or user space)
 * to use the same resource because they can find the address from the name.
*/
	flags = rt_spin_lock_irqsave(&list_lock);
	// index 0 is reserved for the null slot.
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (!lxrt_list[slot].name) {
			lxrt_list[slot].name = name;
			lxrt_list[slot].adr  = adr;
			lxrt_list[slot].tsk  = tsk;
			lxrt_list[slot].type = type;
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_name(unsigned long name)
{
	unsigned long flags;
	int slot;
/*
 * Deregister the resource.
*/
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			// Reset to the null slot.
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline int drg_on_adr(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
			lxrt_list[slot] = lxrt_list[0];
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return slot;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline unsigned long get_name(void *adr)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].adr == adr) {
		rt_spin_unlock_irqrestore(flags, &list_lock);
		return lxrt_list[slot].name;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

static inline void *get_adr(unsigned long name)
{
	unsigned long flags;
	int slot;
	flags = rt_spin_lock_irqsave(&list_lock);
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].name == name) {
			rt_spin_unlock_irqrestore(flags, &list_lock);
			return lxrt_list[slot].adr;
		}
	}
	rt_spin_unlock_irqrestore(flags, &list_lock);
	return 0;
} 

int rt_register(unsigned long name, void *adr)
{
/*
 * Register a resource. This function provides the service to all RTAI tasks.
*/
	return registr(name, adr, IS_TASK, 0 );
}

int rt_drg_on_name(unsigned long name)
{
	return drg_on_name(name);
} 

int rt_drg_on_adr(void *adr)
{
	return drg_on_adr(adr);
} 

unsigned long rt_get_name(void *adr)
{
	return get_name(adr);
} 

void *rt_get_adr(unsigned long name)
{
	return get_adr(name);
}

#define LXRT_RESUME(x)  {extern void emuser_trxl(RT_TASK *); emuser_trxl((x));}
#define LXRT_SUSPEND(x) {extern void (*dnepsus_trxl)(void); (*dnepsus_trxl)();}

// BE CAREFULL TO KEEP THE 3 MACROS BELOW UP TO DATE WITH THE SCHEDULERS!
// WATCH OUT for:    1) Largest stack size required by any scheduler function;
//		     2) Max size of messages expected in mail boxes;
//		     3) Max number of 4 bytes args in scheduler functions.
// Note that the first two are just defaults. You are in full control of the 
// stack size at task init and the mail box size is automatically updated when
// needed, with a GFP_KERNEL non atomic kmalloc. 

#define STACK_SIZE    512	// Default stack size.
#define MSG_SIZE      256	// Default max message size.
#define MAX_ARGS       10	// Max 4 bytes args in rtai functions.

/*       !!! ATTENTION: KEEP THE MACRO BELOW A POWER OF 2, ALWAYS !!! 
         Must be >= the max number of expected real time Linux tasks.       */
#define MAX_SRQ  128

/*
 * For any of MAX_SRQ user space programs, there is no certainty that RTAI
 * will be able to complete the requested service immediately.
 * sysrq is used to queue pending system calls that must be rescheduled
 * by Linux.
 *
 * sigsysrq is used in a similar fashion to send signals asynchronously 
 * to Linux tasks registered with LXRT.
*/

static struct {int srq, in, out; struct wait_queue **waitq[MAX_SRQ]; spinlock_t lock;} sysrq;

#ifdef ENABLE_SIGNAL
static struct {int srq, in, out; struct { struct task_struct *tsk; int sig; } waitq[MAX_SRQ]; spinlock_t lock;} sigsysrq;
#endif

typedef union { unsigned long args[MAX_ARGS]; } ARGS;

static void lxrt_suspend(ARGS args)
{
#define rt_task ((RT_TASK *)(args.args[MAX_ARGS - 1]))
	__asm__ __volatile__ ("" : "=A" (rt_task->retval) );
	rt_global_cli();

	if ((rt_task->lnxtsk.task)->state == TASK_INTERRUPTIBLE) {
		// The Linux system call was suspended.	
		sysrq.waitq[sysrq.in] = &(rt_task->waitq);
		sysrq.in = (sysrq.in + 1) & (MAX_SLOTS - 1);
		rt_pend_linux_srq(sysrq.srq);
	}
	// rt_task returned from the requested fun() and has nothing else to do.
	rt_task->state = 0;
	// The next line calls rt_schedule:
	LXRT_SUSPEND(rt_task);
	printk("!!! GREAT PAIN, THIS POINT CANNOT BE REACHED !!!\n"); 
#undef rt_task
}

static void lxrt_global_sti(void)
{
	if (test_and_clear_bit(hard_cpu_id(), locked_cpus)) {
		clear_bit(31, locked_cpus);
	}
	hard_sti();
}

static inline long long lxrt_resume(void *fun, int narg, int *arg, int type, RT_TASK *rt_task)
{
	int msg_size, *msg_adr;

	// copy the arguments on the stack of the RTAI task
	copy_from_user(rt_task->stack_top, arg, (narg - 1)*sizeof(int));

/* Here type > 0 means any messaging with the need of copying from/to user 
   space. My knowledge of Linux memory managment has led to this mess. 
   Whoever can do it better is warmly welcomed.                             */
	msg_adr = 0;
	msg_size = 0;
	switch( type ) {
		case 0: // rt_send(), rt_sem()
			break;

		case 1: // rt_receive()		
                case 2: // rt_rpc()
                        msg_size = sizeof(int);
                        msg_adr  = (int *)(*(rt_task->stack_top + type));
                        *(rt_task->stack_top + type) = (int)rt_task->msg_buf;
                        break;

                case 3: // rt_mbx_send()
                case 4: // rt_mbx_receive()
                        if ((msg_size = *(rt_task->stack_top + 2)) > rt_task->max_msg_size) {
                                kfree(rt_task->msg_buf);
                                rt_task->max_msg_size = ((*(rt_task->stack_top + 2))*120 + 50)/100;
                                rt_task->msg_buf = kmalloc(rt_task->max_msg_size, GFP_KERNEL);
                        }
                        msg_adr  = (int *)(*(rt_task->stack_top + 1));
                        *(rt_task->stack_top + 1) = (int)rt_task->msg_buf;
                        if( type == 3 ) copy_from_user(rt_task->msg_buf, msg_adr, msg_size);
                        break;
		}

	*(rt_task->stack_top - 1) = (int)lxrt_suspend;
	*(rt_task->stack_top - 2) = (int)fun;
	*(rt_task->stack = rt_task->stack_top - 3) = (int)lxrt_global_sti;
	rt_global_cli();
	// The following line is the backdoor into RTAI context!
	LXRT_RESUME(rt_task);

	rt_global_cli();
	if (rt_task->state) {
		current->state = TASK_INTERRUPTIBLE;
		rt_global_sti();
		// Kick in the Linux scheduler.
		schedule();
	} else {
		rt_global_sti();
		}
/*
 * We may have some data to pick up for the calling user space program.
*/
	if (type > 0 && type != 3) {
		// For rt_mbx_receive, rt_receive and rt_rpc.
		copy_to_user(msg_adr, rt_task->msg_buf, msg_size);
	}

	// On our way back to user space.
	return rt_task->retval;
}

long long lxrt_handler(unsigned int lxsrq, void *arg)
{
	union {unsigned long name; RT_TASK *rt_task; SEM *sem; MBX *mbx; } arg0;
	unsigned long name;
	int srq;

	get_user(arg0.name, (unsigned long *)arg);
	if ((srq = SRQ(lxsrq)) < GET_ADR) {
		if (rt_fun[srq].type) {
			// Deal with the scheduler services. 
			return lxrt_resume(rt_fun[srq].fun, NARG(lxsrq), &((int *)arg)[1], rt_fun[srq].type - 1, arg0.rt_task);
		}

		switch (NARG(lxsrq)) {
			// Other RTAI services with 0, 1 or 2 arguments.
			case 0:
				return ((long long (*)(void))rt_fun[srq].fun)();
			case 1:
				return ((long long (*)(int))rt_fun[srq].fun)(arg0.name);
			case 2: {
				int arg1;
				get_user(arg1, &((int *)arg)[1]);
				return ((long long (*)(int, int))rt_fun[srq].fun)(arg0.name, arg1);
			}
		}
	}

	switch (srq) { // Services specific to LXRT come next.

		case GET_ADR: {
			return (unsigned long)get_adr(arg0.name);
		}

		case GET_NAME: {
			return get_name((void *)arg0.name);
		}

		case TASK_INIT: {
			int prio, stack_size, max_msg_size;
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			get_user(prio, &((int *)arg)[1]);
			if (prio > RT_LOWEST_PRIORITY) {
				prio = RT_LOWEST_PRIORITY;
			}
			get_user(stack_size, &((int *)arg)[2]);
			if (!stack_size) {
				stack_size = STACK_SIZE;
			}
			get_user(max_msg_size, &((int *)arg)[3]);
			if (!max_msg_size) {
				max_msg_size = MSG_SIZE;
			}
			if (!(arg0.rt_task = kmalloc(sizeof(RT_TASK), GFP_KERNEL)) ||
			    !(arg0.rt_task->msg_buf = kmalloc(max_msg_size, GFP_KERNEL))) {
				return 0;
			}
			if (!rt_task_init(arg0.rt_task, 0, 0, stack_size, prio, 0, 0)) {
				*(arg0.rt_task->stack) = (int)arg0.rt_task;
				arg0.rt_task->stack_top = arg0.rt_task->stack - MAX_ARGS;
				arg0.rt_task->max_msg_size = max_msg_size;
				arg0.rt_task->state = 0;
				arg0.rt_task->waitq = 0;
				// This is how the RTAI task becomes aware of the Linux task.
				arg0.rt_task->lnxtsk.task = current;
				__add_wait_queue(&(arg0.rt_task->waitq), &(arg0.rt_task->lnxtsk));
				if (registr(name, arg0.rt_task, IS_TASK, 0)) {
#ifdef VERBOSE
printk("Create RT_TASK %X\n", (int) arg0.rt_task);
#endif
					return arg0.name;
				} else {
					kfree(arg0.rt_task);
				}
			}
			return 0;
		}

		// Delete the RTAI image process.
		case TASK_DELETE: {
			if (rt_task_delete(arg0.rt_task)) {
				return -EFAULT;
			}
			kfree(arg0.rt_task->msg_buf);
			kfree(arg0.rt_task);
			return drg_on_adr(arg0.rt_task);
		}

		case SEM_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.sem = kmalloc(sizeof(SEM), GFP_KERNEL))) {
				int ini_cnt;
				get_user(ini_cnt, &((int *)arg)[1]);
				rt_sem_init(arg0.sem, ini_cnt);
				if (registr(name, arg0.sem, IS_SEM, current)) {
					return arg0.name;
				} else {
					kfree(arg0.sem);
				}
			}
			return 0;
		}

		case SEM_DELETE: {
			if (rt_sem_delete(arg0.sem)) {
				return -EFAULT;
			}
			kfree(arg0.sem);
			return drg_on_adr(arg0.sem);
		}

		case MBX_INIT: {
			if (get_adr(name = arg0.name)) {
				return 0;
			}
			if ((arg0.mbx = kmalloc(sizeof(MBX), GFP_KERNEL))) {
				int size;
				get_user(size, &((int *)arg)[1]);
				rt_mbx_init(arg0.mbx, size);
				if (registr(name, arg0.mbx, IS_MBX, current)) {
					return arg0.name;
				} else {
					kfree(arg0.mbx);
				}
			}
			return 0;
		}

		case MBX_DELETE: {
			if (rt_mbx_delete(arg0.mbx)) {
				return -EFAULT;
			}
			kfree(arg0.mbx);
			return drg_on_adr(arg0.mbx);
		}
	}
	return 0;
}

static void lxrt_srq_handler(void)
{
/*
 * This is a Linux context handler responding to system request from RTAI.
*/  
	while (sysrq.out != sysrq.in) {
		wake_up_interruptible(sysrq.waitq[sysrq.out]);
		sysrq.out = (sysrq.out + 1) & (MAX_SLOTS - 1);
	}
}

#ifdef ENABLE_SIGNAL
static void lxrt_sigsrq_handler(void)
{
/*
 * This is a Linux context handler responding to system request from RTAI.
*/
struct task_struct *tsk ;
int err, sig ;

        while (sigsysrq.out != sigsysrq.in) {
                tsk = sigsysrq.waitq[sigsysrq.out].tsk ;
		sig = sigsysrq.waitq[sigsysrq.out].sig ;
		err = kill_proc(tsk->pid, sig, 0 );

                sigsysrq.out = (sigsysrq.out + 1) & (MAX_SLOTS - 1);
        }
}
#endif


#ifdef ENABLE_SIGNAL_TEST
RT_TASK test_task;

void test_task_code(int notused)
{
//unsigned int err=0;
//RT_TASK *t;

rt_sleep(nano2count(4*1E9));
rt_signal_linux_task(get_adr(nam2num("SRV")), SIGINT);

//t = get_adr(nam2num("SRV"));
//if(t) {
//	err = (int) rt_rpc( t, err, &err );
//	}
//printk("rt_send(%X) returns %d\n", (int) t, err );

rt_task_suspend(&test_task);
}
#endif

static struct desc_struct sidt;

int init_module(void)
{
	// rtai_lxrt_handler() is the handler for soft int RTAI_LXRT_VECTOR
	// 15 -> trap, not a hardware interrupt
	//  3 -> dpl, privilege level, can be called from user space
	sidt = rt_set_full_intr_vect(RTAI_LXRT_VECTOR, 15, 3, rtai_lxrt_handler);
	if ((sysrq.srq = rt_request_srq(nam2num("LXRT"), lxrt_srq_handler, 0)) < 0) {
		printk("LXRT: no sysrq available.\n");
		return sysrq.srq;
		}
	sysrq.in = sysrq.out = sysrq.lock.lock = 0;

#ifdef ENABLE_SIGNAL
        if ((sigsysrq.srq = rt_request_srq(nam2num("LSIG"), lxrt_sigsrq_handler, 0)) < 0) {
                printk("LXRT: no sigsysrq available.\n");
                return sigsysrq.srq;
        	}
        sigsysrq.in = sigsysrq.out = sigsysrq.lock.lock = 0;
#endif

	// Linux will call rtai_callback() whenever a task with SCHED_FIFO
	// or SCHED_RR scheduling policy is aborted.       
	if(set_rtai_callback( rtai_callback )) {
                printk("Could not setup rtai_callback\n");
	        }

#ifdef ENABLE_SIGNAL_TEST
        if(rt_task_init(&test_task, test_task_code, 0, 3000, 10, 0, 0) != 0) {
                printk("Could not start test task\n");
                }
       
        //rt_set_oneshot_mode();
        start_rt_timer(nano2count(1E9));
        rt_task_resume(&test_task);
#endif

return 0 ;
}

void cleanup_module(void)
{

#ifdef ENABLE_SIGNAL_TEST
	rt_task_delete(&test_task);
#endif

	remove_rtai_callback( rtai_callback );

#ifdef ENABLE_SIGNAL
        if (rt_free_srq(sigsysrq.srq) < 0) {
                printk("LXRT: sigsysrq %d illegal or already free.\n", sigsysrq.srq);
        }
#endif

	if (rt_free_srq(sysrq.srq) < 0) {
		printk("LXRT: sysrq %d illegal or already free.\n", sysrq.srq);
	}
	rt_reset_full_intr_vect(RTAI_LXRT_VECTOR, sidt);
	return;
}

static RT_TASK *base_linux_tasks[NR_RT_CPUS];

static RT_TASK *my_get_base_linux_task(int cpu)
{
int stype;

rt_get_base_linux_task(base_linux_tasks);

stype = rt_sched_type();
if( stype == UP_SCHED || stype == SMP_SCHED ) {
	return cpu ? 0 : base_linux_tasks[0];
	}
else if( stype == MUP_SCHED ) {
	return base_linux_tasks[cpu];
        }
return 0; // To keep compiler happy
}

void rtai_callback( struct task_struct *ltsk )
{
RT_TASK *task2delete, *task2unblock ;
int cpu, slot;
char name[8];

/*
 * Linux is just about to schedule *ltsk out of existence.
 * With this feature, LXRT becomes an informed administrator a la QNX.
*/
rt_global_cli();

for( cpu=0 ; cpu < NR_RT_CPUS ; cpu++ ) {

	// Only do more than one cpu if using MUP_SCHED
	if(cpu && (rt_sched_type() != MUP_SCHED)) break;

	task2delete = my_get_base_linux_task(cpu);
	if(!task2delete) continue ; // This is not suppose to happen

	while ((task2delete = task2delete->next)) {
		// Try to find if RTAI was aware of this dying Linux task.	
		if (task2delete->lnxtsk.task == ltsk ) break;
		}

	// First let's free the registered resources.
	for (slot = 1; slot <= MAX_SLOTS; slot++) {
		if (lxrt_list[slot].tsk == ltsk ) {
			switch(lxrt_list[slot].type) {
				case IS_SEM:
					rt_sem_delete(lxrt_list[slot].adr); 
#ifdef VERBOSE
printk("Delete SEM %X\n", (int) lxrt_list[slot].adr);
#endif
          			        kfree(lxrt_list[slot].adr);
					break;
	                        case IS_MBX:
#ifdef VERBOSE
printk("Delete MBX %X\n", (int) lxrt_list[slot].adr);
#endif
        	                        rt_mbx_delete(lxrt_list[slot].adr);
                	                kfree(lxrt_list[slot].adr);
					break;

// to do:			case IS_SHMEM:
				 
				}
	                lxrt_list[slot] = lxrt_list[0];
			}
		}

        if(!task2delete) {
                continue; // The user deleted the task but forgot to delete the resources.
                }

	// Other RTAI tasks may be SEND, RPC or RETURN blocked on task2delete.
	task2unblock = my_get_base_linux_task(cpu);
	while ((task2unblock = task2unblock->next)) {
	        if (!(task2unblock->state & (SEND | RPC | RETURN))) continue ;
        	else if(task2unblock->msg_queue.task == task2delete) {
             	        task2unblock->state &= ~(SEND | RPC | RETURN | DELAYED);
                	LXRT_RESUME(task2unblock);
	                }
	        }

	// To do: other RTAI tasks may want to be informed as well.

	// Ok, let's delete the task.
#ifdef VERBOSE
num2nam( get_name(task2delete), name );
printk( "Dump RTAI task %s\n", name );
#endif
	if (!rt_task_delete(task2delete)) {
	        kfree(task2delete->msg_buf);
        	kfree(task2delete);
       		drg_on_adr(task2delete);
		break;
        	}
	}

rt_global_sti();
}

#ifdef ENABLE_SIGNAL
int rt_signal_linux_task(RT_TASK *task, int sig)
{
/*
 * A RTAI task wants to send a signal asynchronously to a Linux task.
 * LXRT does not need this function. It's here because I wanted to experiment
 * with signals. Could be used by an interrupt handler to notify a Linux
 * task that data is available with say SIGUSR1.
*/

if(!task || !task->lnxtsk.task) return -EFAULT;	

// Setup the srq to signal the Linux counterpart of the real time task tsk
rt_global_cli();
sigsysrq.waitq[sigsysrq.in].tsk = task->lnxtsk.task;
sigsysrq.waitq[sigsysrq.in].sig = sig ;
sigsysrq.in = (sigsysrq.in + 1) & (MAX_SLOTS - 1);
rt_global_sti();

rt_pend_linux_srq(sigsysrq.srq);

return 0;
}
#endif
